--------------------------------------------------------------------
-- $Id: QuickCheckBatch.hs,v 1.1.2.2 2000/01/21 12:23:53 simonmar Exp $

-- This is a batch driver for runing QuickCheck.
-- Andy Gill

module QuickCheckBatch
   ( run
   , runTests
   ) where

import Random
import QuickCheck

data TestResult = TestOk 	String  Int [[String]]
		| TestExausted 	String  Int [[String]]
		| TestFailed   [String] Int

tests :: Config -> Gen Result -> StdGen -> Int -> Int -> [[String]] 
      -> IO TestResult
tests config gen rnd0 ntest nfail stamps
  | ntest == maxTest config = return (TestOk  "OK, passed" ntest stamps)
  | nfail == maxFail config = return (TestExausted "Arguments exhausted after"
					 ntest stamps)
  | otherwise               =
      do case ok result of
           Nothing    ->
             tests config gen rnd1 ntest (nfail+1) stamps
           Just True  ->
             tests config gen rnd1 (ntest+1) nfail (stamp result:stamps)
           Just False ->
             do return (TestFailed (arguments result) ntest)
     where
      result      = generate (size config ntest) rnd2 gen
      (rnd1,rnd2) = split rnd0

batch n = Config
  { maxTest = n
  , maxFail = n * 10
  , size    = (+ 3) . (`div` 2)
  , every   = \n args -> ""
  }

-- Here we use the same random number each time,
-- so we get reproducable results!
run :: Testable a => a -> Int -> IO TestResult
run a n = tests (batch n) (evaluate a) (mkStdGen 0) 0 0 []

-- Prints a one line summary of various tests with common theme
runTests :: String -> Int -> [Int -> IO TestResult] -> IO ()
runTests name scale actions =
  do putStr (rjustify 25 name ++ " : ")
     f <- tr 1 actions [] 0
     mapM fa f
     return ()
  where
	rjustify n s = replicate (max 0 (n - length s)) ' ' ++ s

	tr n [] xs c = do { putStr ("\n")  ; return xs }
	tr n (action:actions) others c = 
	   do r <- action scale
	      case r of
		(TestOk _ m _) 
			-> do { putStr "." ;
			       tr (n+1) actions others (c+m) }
		(TestExausted _ m _) 
			-> do { putStr ":" ;
			       tr (n+1) actions others (c+m) }
	  	(TestFailed f num)
			-> do { putStr "#" ;
			        tr (n+1) actions ((f,n,num):others) (c+num) }

	fa :: ([String],Int,Int) -> IO ()
	fa (f,n,no) = 
	  do putStr "\n"
	     putStr ("    ** test " 
			++ show (n  :: Int)
			++ " of "
			++ name
			++ " failed with the binding(s)\n")
	     sequence_ [putStr ("    **   " ++ v ++ "\n")
			| v <- f ]
  	     putStr "\n"

{-
 - Here is is an example of use for sorting:
 - 
 - 
 - prop_sort1 xs = sort xs == sortBy compare xs
 -   where types = (xs :: [OrdALPHA])
 - prop_sort2 xs =
 - 	(not (null xs)) ==>
 - 	(head (sort xs) == minimum xs)
 -   where types = (xs :: [OrdALPHA])
 - prop_sort3 xs = 
 -      (not (null xs)) ==>
 - 	last (sort xs) == maximum xs
 -   where types = (xs :: [OrdALPHA])
 - prop_sort4 xs ys =
 - 	(not (null xs)) ==>
 - 	(not (null ys)) ==>
 - 	(head (sort (xs ++ ys)) == min (minimum xs) (minimum ys))
 -   where types = (xs :: [OrdALPHA], ys :: [OrdALPHA])
 - prop_sort5 xs ys =
 - 	(not (null xs)) ==>
 - 	(not (null ys)) ==>
 - 	(last (sort (xs ++ ys)) == max (maximum xs) (maximum ys))
 -   where types = (xs :: [OrdALPHA], ys :: [OrdALPHA])
 - 
 - check_sort :: IO ()
 - check_sort = runTests "sort" 100
 - 	[ run prop_sort1
 -  	, run prop_sort2
 -  	, run prop_sort3
 -  	, run prop_sort4
 -  	, run prop_sort5
 - 	]
 - 
 - When run, this gives
 - Main> check_sort
 -                     sort : .....
 - 
 - You would tie together all the check_* functions
 - into one check_everything.
 -
 -}
